@@ -4,7 +4,6 @@ from __future__ import annotations
 
 import asyncio
 import base64
-import binascii
 import gzip
 import hashlib
 import hmac
@@ -14,17 +13,18 @@ import math
 import secrets
 import struct
 import time
+from random import randint
 from typing import Any, Callable
 
 import aiohttp
 from Crypto.Cipher import AES
-from Crypto.Util.Padding import pad, unpad
+from Crypto.Util.Padding import unpad
 
 from roborock.exceptions import (
     RoborockException, RoborockTimeout, VacuumError,
 )
-from .code_mappings import WASH_MODE_MAP, DUST_COLLECTION_MAP, RoborockDockType, \
-    RoborockDockDustCollectionType, RoborockDockWashingModeType, STATE_CODE_TO_STATUS
+from .code_mappings import RoborockDockType, \
+    STATE_CODE_TO_STATUS
 from .containers import (
     UserData,
     Status,
@@ -34,15 +34,19 @@ from .containers import (
     CleanRecord,
     HomeData,
     MultiMapsList,
-    SmartWashParameters,
+    SmartWashParams,
+    RoborockDeviceInfo,
+    WashTowelMode,
+    DustCollectionMode,
+    NetworkInfo,
 
 )
+from .roborock_message import RoborockMessage
 from .roborock_queue import RoborockQueue
 from .typing import (
     RoborockDeviceProp,
     RoborockCommand,
-    RoborockDockSummary,
-    RoborockDeviceInfo,
+    RoborockDockSummary
 )
 
 _LOGGER = logging.getLogger(__name__)
@@ -59,17 +63,6 @@ def md5hex(message: str) -> str:
     return md5.hexdigest()
 
 
-def md5bin(message: str) -> bytes:
-    md5 = hashlib.md5()
-    md5.update(message.encode())
-    return md5.digest()
-
-
-def encode_timestamp(_timestamp: int) -> str:
-    hex_value = f"{_timestamp:x}".zfill(8)
-    return "".join(list(map(lambda idx: hex_value[idx], [5, 6, 3, 7, 1, 2, 0, 4])))
-
-
 class PreparedRequest:
     def __init__(self, base_url: str, base_headers: dict = None) -> None:
         self.base_url = base_url
@@ -95,143 +88,73 @@ class RoborockClient:
 
     def __init__(self, endpoint: str, devices_info: dict[str, RoborockDeviceInfo]) -> None:
         self.devices_info = devices_info
-        self._seq = 1
-        self._random = 4711
-        self._id_counter = 10000
         self._salt = "TXdfu$jyZ#TZHsg4"
         self._endpoint = endpoint
         self._nonce = secrets.token_bytes(16)
         self._waiting_queue: dict[int, RoborockQueue] = {}
-        self._status_listeners: list[Callable[[str, str], None]] = []
+        self._status_listeners: list[Callable[[int, str], None]] = []
 
-    def add_status_listener(self, callback: Callable[[str, str], None]):
+    def add_status_listener(self, callback: Callable[[int, str], None]):
         self._status_listeners.append(callback)
 
     async def async_disconnect(self) -> Any:
         raise NotImplementedError
 
-    def _decode_msg(self, msg: bytes, local_key: str) -> dict[str, Any]:
-        if msg[4:7] == "1.0".encode():
-            msg = msg[4:]
-        elif msg[0:3] != "1.0".encode():
-            raise RoborockException(f"Unknown protocol version {msg[0:3]}")
-        if len(msg) == 17:
-            [version, request_id, _random, timestamp, protocol] = struct.unpack(
-                "!3sIIIH", msg[0:17]
-            )
-            return {
-                "version": version,
-                "request_id": request_id,
-                "timestamp": timestamp,
-                "protocol": protocol,
-            }
-        [version, request_id, _random, timestamp, protocol, payload_len] = struct.unpack(
-            "!3sIIIHH", msg[0:19]
-        )
-        extra_len = len(msg) - 23 - payload_len
-        [payload, expected_crc32, extra] = struct.unpack_from(f"!{payload_len}sI{extra_len}s", msg, 19)
-        if not extra_len:
-            crc32 = binascii.crc32(msg[0: 19 + payload_len])
-            if crc32 != expected_crc32:
-                raise RoborockException(f"Wrong CRC32 {crc32}, expected {expected_crc32}")
-
-        aes_key = md5bin(encode_timestamp(timestamp) + local_key + self._salt)
-        decipher = AES.new(aes_key, AES.MODE_ECB)
-        decrypted_payload = unpad(decipher.decrypt(payload), AES.block_size) if payload else extra
-        return {
-            "version": version,
-            "request_id": request_id,
-            "timestamp": timestamp,
-            "protocol": protocol,
-            "payload": decrypted_payload
-        }
-
-    def _encode_msg(self, device_id, request_id, protocol, timestamp, payload, prefix=None) -> bytes:
-        local_key = self.devices_info[device_id].device.local_key
-        aes_key = md5bin(encode_timestamp(timestamp) + local_key + self._salt)
-        cipher = AES.new(aes_key, AES.MODE_ECB)
-        encrypted = cipher.encrypt(pad(payload, AES.block_size))
-        encrypted_len = len(encrypted)
-        values = [
-            "1.0".encode(),
-            request_id,
-            self._random,
-            timestamp,
-            protocol,
-            encrypted_len,
-            encrypted
-        ]
-        if prefix:
-            values = [prefix] + values
-        msg = struct.pack(
-            f"!{'I' if prefix else ''}3sIIIHH{encrypted_len}s",
-            *values
-        )
-        crc32 = binascii.crc32(msg[4:] if prefix else msg)
-        msg += struct.pack("!I", crc32)
-        return msg
-
-    async def on_message(self, device_id, msg) -> bool:
+    async def on_message(self, messages: list[RoborockMessage]) -> None:
         try:
-            data = self._decode_msg(msg, self.devices_info[device_id].device.local_key)
-            protocol = data.get("protocol")
-            if protocol == 102 or protocol == 4:
-                payload = json.loads(data.get("payload").decode())
-                for data_point_number, data_point in payload.get("dps").items():
-                    if data_point_number == "102":
-                        data_point_response = json.loads(data_point)
-                        request_id = data_point_response.get("id")
+            for data in messages:
+                protocol = data.protocol
+                if protocol == 102 or protocol == 4:
+                    payload = json.loads(data.payload.decode())
+                    for data_point_number, data_point in payload.get("dps").items():
+                        if data_point_number == "102":
+                            data_point_response = json.loads(data_point)
+                            request_id = data_point_response.get("id")
+                            queue = self._waiting_queue.get(request_id)
+                            if queue:
+                                if queue.protocol == protocol:
+                                    error = data_point_response.get("error")
+                                    if error:
+                                        await queue.async_put(
+                                            (
+                                                None,
+                                                VacuumError(
+                                                    error.get("code"), error.get("message")
+                                                ),
+                                            ),
+                                            timeout=QUEUE_TIMEOUT,
+                                        )
+                                    else:
+                                        result = data_point_response.get("result")
+                                        if isinstance(result, list) and len(result) == 1:
+                                            result = result[0]
+                                        await queue.async_put(
+                                            (result, None), timeout=QUEUE_TIMEOUT
+                                        )
+                        elif data_point_number == "121":
+                            status = STATE_CODE_TO_STATUS.get(data_point)
+                            _LOGGER.debug(f"Status updated to {status}")
+                            for listener in self._status_listeners:
+                                listener(data.seq, status)
+                        else:
+                            _LOGGER.debug(
+                                f"Unknown data point number received {data_point_number} with {data_point}"
+                            )
+                elif protocol == 301:
+                    payload = data.payload[0:24]
+                    [endpoint, _, request_id, _] = struct.unpack("<15sBH6s", payload)
+                    if endpoint.decode().startswith(self._endpoint):
+                        iv = bytes(AES.block_size)
+                        decipher = AES.new(self._nonce, AES.MODE_CBC, iv)
+                        decrypted = unpad(
+                            decipher.decrypt(data.payload[24:]), AES.block_size
+                        )
+                        decrypted = gzip.decompress(decrypted)
                         queue = self._waiting_queue.get(request_id)
                         if queue:
-                            if queue.protocol == protocol:
-                                error = data_point_response.get("error")
-                                if error:
-                                    await queue.async_put(
-                                        (
-                                            None,
-                                            VacuumError(
-                                                error.get("code"), error.get("message")
-                                            ),
-                                        ),
-                                        timeout=QUEUE_TIMEOUT,
-                                    )
-                                else:
-                                    result = data_point_response.get("result")
-                                    if isinstance(result, list) and len(result) == 1:
-                                        result = result[0]
-                                    await queue.async_put(
-                                        (result, None), timeout=QUEUE_TIMEOUT
-                                    )
-                                    return True
-                        elif request_id < self._id_counter:
-                            _LOGGER.debug(
-                                f"id={request_id} Ignoring response: {data_point_response}"
-                            )
-                    elif data_point_number == "121":
-                        status = STATE_CODE_TO_STATUS.get(data_point)
-                        _LOGGER.debug(f"Status updated to {status}")
-                        for listener in self._status_listeners:
-                            listener(device_id, status)
-                    else:
-                        _LOGGER.debug(
-                            f"Unknown data point number received {data_point_number} with {data_point}"
-                        )
-            elif protocol == 301:
-                payload = data.get("payload")[0:24]
-                [endpoint, _, request_id, _] = struct.unpack("<15sBH6s", payload)
-                if endpoint.decode().startswith(self._endpoint):
-                    iv = bytes(AES.block_size)
-                    decipher = AES.new(self._nonce, AES.MODE_CBC, iv)
-                    decrypted = unpad(
-                        decipher.decrypt(data.get("payload")[24:]), AES.block_size
-                    )
-                    decrypted = gzip.decompress(decrypted)
-                    queue = self._waiting_queue.get(request_id)
-                    if queue:
-                        if isinstance(decrypted, list):
-                            decrypted = decrypted[0]
-                        await queue.async_put((decrypted, None), timeout=QUEUE_TIMEOUT)
-                        return True
+                            if isinstance(decrypted, list):
+                                decrypted = decrypted[0]
+                            await queue.async_put((decrypted, None), timeout=QUEUE_TIMEOUT)
         except Exception as ex:
             _LOGGER.exception(ex)
 
@@ -252,8 +175,7 @@ class RoborockClient:
             self, method: RoborockCommand, params: list = None, secured=False
     ):
         timestamp = math.floor(time.time())
-        request_id = self._id_counter
-        self._id_counter += 1
+        request_id = randint(10000, 99999)
         inner = {
             "id": request_id,
             "method": method,
@@ -262,13 +184,13 @@ class RoborockClient:
         if secured:
             inner["security"] = {
                 "endpoint": self._endpoint,
-                "nonce": self._nonce.hex().upper(),
+                "nonce": self._nonce.hex().lower(),
             }
         payload = bytes(
             json.dumps(
                 {
-                    "t": timestamp,
                     "dps": {"101": json.dumps(inner, separators=(",", ":"))},
+                    "t": timestamp,
                 },
                 separators=(",", ":"),
             ).encode()
@@ -323,25 +245,27 @@ class RoborockClient:
         except RoborockTimeout as e:
             _LOGGER.error(e)
 
-    async def get_washing_mode(self, device_id: str) -> RoborockDockWashingModeType:
+    async def get_wash_towel_mode(self, device_id: str) -> WashTowelMode:
         try:
             washing_mode = await self.send_command(device_id, RoborockCommand.GET_WASH_TOWEL_MODE)
-            return WASH_MODE_MAP.get(washing_mode['wash_mode'])
+            if isinstance(washing_mode, dict):
+                return WashTowelMode(washing_mode)
         except RoborockTimeout as e:
             _LOGGER.error(e)
 
-    async def get_dust_collection_mode(self, device_id: str) -> RoborockDockDustCollectionType:
+    async def get_dust_collection_mode(self, device_id: str) -> DustCollectionMode:
         try:
             dust_collection = await self.send_command(device_id, RoborockCommand.GET_DUST_COLLECTION_MODE)
-            return DUST_COLLECTION_MAP.get(dust_collection['mode'])
+            if isinstance(dust_collection, dict):
+                return DustCollectionMode(dust_collection)
         except RoborockTimeout as e:
             _LOGGER.error(e)
 
-    async def get_mop_wash_mode(self, device_id: str) -> SmartWashParameters:
+    async def get_smart_wash_params(self, device_id: str) -> SmartWashParams:
         try:
             mop_wash_mode = await self.send_command(device_id, RoborockCommand.GET_SMART_WASH_PARAMS)
             if isinstance(mop_wash_mode, dict):
-                return SmartWashParameters(mop_wash_mode)
+                return SmartWashParams(mop_wash_mode)
         except RoborockTimeout as e:
             _LOGGER.error(e)
 
@@ -349,14 +273,21 @@ class RoborockClient:
         try:
             commands = [self.get_dust_collection_mode(device_id)]
             if dock_type == RoborockDockType.EMPTY_WASH_FILL_DOCK:
-                commands += [self.get_mop_wash_mode(device_id), self.get_washing_mode(device_id)]
-            [collection_mode, mop_wash, washing_mode] = (list(await asyncio.gather(*commands)) + [None, None])[:3]
+                commands += [self.get_wash_towel_mode(device_id), self.get_smart_wash_params(device_id)]
+            [
+                dust_collection_mode,
+                wash_towel_mode,
+                smart_wash_params
+            ] = (
+                        list(await asyncio.gather(*commands))
+                        + [None, None]
+                )[:3]
 
-            return RoborockDockSummary(collection_mode, washing_mode, mop_wash)
+            return RoborockDockSummary(dust_collection_mode, wash_towel_mode, smart_wash_params)
         except RoborockTimeout as e:
             _LOGGER.error(e)
 
-    async def get_prop(self, device_id: str) -> RoborockDeviceProp:
+    async def get_prop(self, device_id: str) -> RoborockDeviceProp | None:
         [status, dnd_timer, clean_summary, consumable] = await asyncio.gather(
             *[
                 self.get_status(device_id),
@@ -365,21 +296,41 @@ class RoborockClient:
                 self.get_consumable(device_id),
             ]
         )
+        last_clean_record = None
+        if clean_summary and clean_summary.records and len(clean_summary.records) > 0:
+            last_clean_record = await self.get_clean_record(
+                device_id, clean_summary.records[0]
+            )
+        dock_summary = None
+        if status and status.dock_type != RoborockDockType.NO_DOCK:
+            dock_summary = await self.get_dock_summary(device_id, status.dock_type)
         if any([status, dnd_timer, clean_summary, consumable]):
             return RoborockDeviceProp(
-                status, dnd_timer, clean_summary, consumable
+                status, dnd_timer, clean_summary, consumable, last_clean_record, dock_summary
             )
+        return None
 
     async def get_multi_maps_list(self, device_id) -> MultiMapsList:
-        multi_maps_list = await self.send_command(
-            device_id, RoborockCommand.GET_MULTI_MAPS_LIST
-        )
-        if isinstance(multi_maps_list, dict):
-            return MultiMapsList(multi_maps_list)
+        try:
+            multi_maps_list = await self.send_command(
+                device_id, RoborockCommand.GET_MULTI_MAPS_LIST
+            )
+            if isinstance(multi_maps_list, dict):
+                return MultiMapsList(multi_maps_list)
+        except RoborockTimeout as e:
+            _LOGGER.error(e)
 
     async def get_map_v1(self, device_id):
         return await self.send_command(device_id, RoborockCommand.GET_MAP_V1)
 
+    async def get_networking(self, device_id) -> NetworkInfo:
+        try:
+            networking_info = await self.send_command(device_id, RoborockCommand.GET_NETWORK_INFO)
+            if isinstance(networking_info, dict):
+                return NetworkInfo(networking_info)
+        except RoborockTimeout as e:
+            _LOGGER.error(e)
+
 
 class RoborockApiClient:
     def __init__(self, username: str, base_url=None) -> None:
